from contextlib import redirect_stdout
from io import StringIO
import logging
from pathlib import Path
from tempfile import TemporaryDirectory, NamedTemporaryFile
from clibato import Clibato
from .support import TestCase


class TestClibato(TestCase):
    """Test Clibato"""

    def test_parse_args_action(self):
        """.parse_args() can parse all possible actions"""
        actions = ['init', 'backup', 'restore', 'version']
        for action in actions:
            args = Clibato.parse_args([action])
            self.assertEqual(action, args.action)

    def test_parse_args_reads_arg_verbose(self):
        """.parse_args() understands the --verbose and -v arguments"""
        args = Clibato.parse_args(['version', '--verbose'])
        self.assertEqual(1, args.verbose)

        args = Clibato.parse_args(['version', '-v'])
        self.assertEqual(1, args.verbose)

        args = Clibato.parse_args(['version', '-vv'])
        self.assertEqual(2, args.verbose)

    def test_parse_args_reads_arg_config(self):
        """.parse_args() understands the --config and -c arguments"""
        config_path = Path('~', '.clibato.yml')

        args = Clibato.parse_args(['init', '--config', str(config_path)])
        self.assertEqual(config_path, args.config_path)

        args = Clibato.parse_args(['init', '-c', str(config_path)])
        self.assertEqual(config_path, args.config_path)

    def test_config_file_not_found(self):
        """.execute() shows error when config file can't be located"""
        config_path = 'missing.config.yml'

        with self.assertLogs('clibato', logging.ERROR) as cm:
            app = Clibato()
            result = app.execute(['backup', '-c', str(config_path)])

        self.assertFalse(result)
        self.assert_length(cm.records, 1)
        self.assert_log_record(
            cm.records[0],
            level='ERROR',
            message=f'Configuration not found: {config_path}'
        )

    def test_init(self):
        """Test: clibato init -c /path/to/.clibato.yml"""
        config_dir = TemporaryDirectory()
        config_path = Path(config_dir.name, '.clibato.yml')

        stdout = StringIO()
        with redirect_stdout(stdout) as output:
            app = Clibato()
            app.execute(['init', '-c', str(config_path)])

        expected = '\n'.join([
            f'Configuration created: {config_path.resolve()}',
            '',
            'Modify the file as per your requirements.',
            'Once done, you can run the following commands.',
            '',
            'clibato backup: Perform a backup.',
            'clibato restore: Restore the last backup.',
            ''
        ])

        self.assertEqual(expected, output.getvalue())

    def test_init_config_file_already_exists(self):
        """Test: clibato init logs error if config file already exists"""
        config_file = NamedTemporaryFile(suffix='.clibato.yml')
        config_path = Path(config_file.name).resolve()

        with self.assertLogs('clibato', logging.ERROR) as cm:
            app = Clibato()
            app.execute(['init', '-c', config_file.name])

        self.assert_length(cm.records, 1)
        self.assert_log_record(
            cm.records[0],
            message=f'Configuration already exists: {config_path}',
            level='ERROR'
        )

    def test_backup(self):
        """Test: clibato backup -v -c /path/to/config.yml"""
        source_path, backup_path = self.create_file_fixtures(location='source')
        config_path = self.create_clibato_config({
            'contents': {
                self.BUNNY_PATH: str(source_path / self.BUNNY_PATH),
                self.WABBIT_PATH: str(source_path / self.WABBIT_PATH)
            },
            'destination': {
                'type': 'directory',
                'path': str(backup_path)
            }
        })

        with self.assertLogs('clibato', logging.INFO) as cm, redirect_stdout(StringIO()) as output:
            app = Clibato()
            app.execute(['backup', '-v', '-c', config_path])

        self.assert_length(cm.records, 3)
        self.assert_log_record(
            cm.records[0],
            level='INFO',
            message=f'Loading configuration: {config_path}'
        )

        self.assert_file_contents(backup_path / self.BUNNY_PATH, 'I am a bunny')
        self.assert_file_contents(backup_path / self.WABBIT_PATH, 'I am a wabbit')

        self.assert_output('Backup completed.\n', output.getvalue())

    def test_restore(self):
        """Test: clibato restore -v -c /path/to/config.yml"""
        source_path, backup_path = self.create_file_fixtures(location='backup')
        config_path = self.create_clibato_config({
            'contents': {
                self.BUNNY_PATH: str(source_path / self.BUNNY_PATH),
                self.WABBIT_PATH: str(source_path / self.WABBIT_PATH)
            },
            'destination': {
                'type': 'directory',
                'path': str(backup_path)
            }
        })

        with self.assertLogs('clibato', logging.INFO) as cm, redirect_stdout(StringIO()) as output:
            app = Clibato()
            app.execute(['restore', '-v', '-c', config_path])

        self.assert_length(cm.records, 3)
        self.assert_log_record(
            cm.records[0],
            level='INFO',
            message=f'Loading configuration: {config_path}'
        )

        self.assert_file_contents(source_path / self.BUNNY_PATH, 'I am a bunny')
        self.assert_file_contents(source_path / self.WABBIT_PATH, 'I am a wabbit')

        expected = '\n'.join(['Restore completed.', ''])
        self.assert_output(expected, output.getvalue())

    def test_version(self):
        """Test: clibato version"""
        with redirect_stdout(StringIO()) as output:
            app = Clibato()
            app.execute(['version'])

        self.assert_output(f'Clibato v{Clibato.VERSION}\n', output.getvalue())

    def test_version_verbose(self):
        """Test: clibato version --verbose"""
        with redirect_stdout(StringIO()) as output:
            app = Clibato()
            app.execute(['version', '-v'])

        expected = '\n'.join([
            f'Clibato v{Clibato.VERSION}',
            'Author: Jigarius | jigarius.com',
            'GitHub: github.com/jigarius/clibato',
            ''
        ])

        self.assert_output(expected, output.getvalue())



from shutil import copyfile
from pathlib import Path
import tempfile

from clibato import Clibato, Content, Directory, Config, ConfigError
from .support import TestCase


class TestConfig(TestCase):
    """Test clibato.Config"""

    _FIXTURE_PATH = Path(__file__).parent / 'fixtures' / 'clibato.test.yml'

    def test__eq__(self):
        """.__eq__()"""
        subject = Config(
            contents=[Content('.bashrc'), Content('.zshrc')],
            destination=Directory(path=tempfile.gettempdir())
        )

        self.assertEqual(
            subject,
            Config(
                contents=[Content('.bashrc'), Content('.zshrc')],
                destination=Directory(path=tempfile.gettempdir())
            )
        )

        # Difference in contents
        self.assertNotEqual(
            subject,
            Config(
                contents=[Content('.bashrc'), Content('.todo.txt')],
                destination=Directory(path=tempfile.gettempdir())
            ),
        )

        # Difference in destination
        tempdir = tempfile.TemporaryDirectory()
        self.assertNotEqual(
            subject,
            Config(
                contents=[Content('.bashrc'), Content('.zshrc')],
                destination=Directory(path=tempdir.name)
            )
        )

    def test_from_dict(self):
        """.from_dict() works with a valid dict"""
        self.assertEqual(
            TestConfig._build_config(),
            Config.from_dict({
                'contents': {
                    '.bashrc': None,
                    '.zshrc': None
                },
                'destination': {
                    'type': 'directory',
                    'path': tempfile.gettempdir()
                }
            })
        )

    def test_from_dict_cannot_contain_illegal_keys(self):
        """.from_dict() fails if when extra keys are found"""
        message = 'Config has illegal keys: bar, foo'
        with self.assertRaisesRegex(ConfigError, message):
            Config.from_dict({
                'foo': 'bunny',
                'bar': 'wabbit',
                'contents': {
                    '.bashrc': None
                },
                'destination': {
                    'type': 'directory',
                    'path': tempfile.gettempdir()
                }
            })

    def test_from_dict_cannot_have_missing_keys(self):
        """.from_dict() fails if required keys are missing"""
        message = 'Config has missing keys: contents, destination'
        with self.assertRaisesRegex(ConfigError, message):
            Config.from_dict({})

    def test_from_dict_values_must_be_dictionary(self):
        """.from_dict() fails if values are not dictionaries"""
        for key in ['contents', 'destination']:
            data = {
                'contents': {
                    '.bashrc': None
                },
                'destination': {
                    'type': 'directory',
                    'path': tempfile.gettempdir()
                }
            }
            data[key] = 'oops'

            message = f'Config has illegal value for: {key}'
            with self.assertRaisesRegex(ConfigError, message):
                Config.from_dict(data)

    def test_from_file(self):
        """.from_file()"""
        config_path = self.create_clibato_config({
            'contents': {
                '.bashrc': None,
                '.vimrc': None,
            },
            'destination': {
                'type': 'directory',
                'path': tempfile.gettempdir()
            }
        })

        with self.assertLogs('clibato') as cm:
            self.assertEqual(
                Config(
                    contents=[Content('.bashrc'), Content('.vimrc')],
                    destination=Directory(path=tempfile.gettempdir()),
                ),
                Config.from_file(config_path)
            )

        self.assert_length(cm.records, 1)
        self.assert_log_record(
            cm.records[0],
            level='INFO',
            message=f'Loading configuration: {config_path}'
        )

    def test_from_file_with_non_existent_file(self):
        """.from_file() fails for if file doesn't exist"""
        with self.assertRaises(FileNotFoundError):
            config_file = tempfile.NamedTemporaryFile(suffix='.clibato.yml')
            config_file.close()

            self.assert_file_not_exists(config_file.name)
            self.assertEqual(
                TestConfig._build_config(),
                Config.from_file(config_file.name)
            )

    def test_locate_with_absolute_path(self):
        """.locate() can detect config with absolute paths"""
        self.assertEqual(
            self._FIXTURE_PATH,
            Config.locate(self._FIXTURE_PATH)
        )

    def test_locate_with_relative_path(self):
        """.locate() can detect config with relative paths"""
        with self.chdir(Clibato.ROOT / 'test'):
            self.assertEqual(
                self._FIXTURE_PATH,
                Config.locate(Path('fixtures', 'clibato.test.yml'))
            )

    def test_locate_with_home_path(self):
        """.locate() can detect config with ~/ paths"""
        path = Path.home() / 'clibato.test.yml'
        copyfile(self._FIXTURE_PATH, path)

        self.assertEqual(
            path,
            Config.locate(Path('clibato.test.yml'))
        )

        path.unlink()

    def test_locate_with_non_existent_file(self):
        """.locate() returns None if config file is not found"""
        self.assertIsNone(Config.locate(Path('tmp', '.clibato.yml')))

    def test_contents(self):
        """.contents()"""
        self.assertEqual(
            TestConfig._build_config().contents(),
            [
                Content('.bashrc'),
                Content('.zshrc')
            ]
        )

    def test_destination(self):
        """.destination()"""
        self.assertEqual(
            Directory(tempfile.gettempdir()),
            TestConfig._build_config().destination()
        )

    @staticmethod
    def _build_config(contents=None, destination=None):
        contents = contents or [
            Content('.bashrc'),
            Content('.zshrc'),
        ]
        destination = destination or Directory(path=tempfile.gettempdir())

        return Config(contents, destination)



from pathlib import Path
from tempfile import gettempdir
import unittest

from clibato import Content, ConfigError, Destination, Directory, Repository
from .support import TestCase


class TestDestination(unittest.TestCase):
    """Test destination.Destination"""

    def test_new(self):
        """Destination cannot be instantiated."""
        message = f'Class not instantiable: {Destination.__name__}'
        with self.assertRaisesRegex(NotImplementedError, message):
            Destination()

    def test_from_dict(self):
        """.from_dict() works with a valid dict"""
        subject = Destination.from_dict({
            'type': 'directory',
            'path': gettempdir()
        })

        self.assertEqual(
            subject,
            Directory(path=gettempdir())
        )

    def test_from_dict_with_illegal_type(self):
        """.from_dict() fails with an invalid dict"""
        with self.assertRaisesRegex(ConfigError, 'Illegal type: foobar'):
            Destination.from_dict({'type': 'foobar'})

    def test_from_dict_with_arg_mismatch(self):
        """.from_dict() fails with on argument mismatch"""
        with self.assertRaises(ConfigError) as cm:
            Destination.from_dict({
                'type': 'directory',
                'remote': 'git@github.com:jigarius/clibato.git'
            })

        self.assertEqual(
            "__init__() got an unexpected keyword argument 'remote'",
            str(cm.exception)
        )


class TestDirectory(TestCase):
    """Test destination.Directory"""

    def test_new(self):
        """Instance creation."""
        subject = Directory(path=gettempdir())

        self.assertIsInstance(subject, Directory)

    def test__eq__(self):
        """__eq__()"""
        subject = Directory(path=gettempdir())

        self.assertEqual(subject, Directory(path=gettempdir()))

        self.assertNotEqual(subject, Directory(path=str(Path.home())))

        self.assertNotEqual(
            subject,
            Repository(gettempdir(), 'git@github.com:bunny/wabbit.git')
        )

    def test_inheritance(self):
        """Directory must extend Destination"""
        self.assert_is_subclass(Directory, Destination)

    def test_path(self):
        """.path()"""
        subject = Directory(gettempdir())

        self.assertEqual(Path(gettempdir()), subject.path())

    def test_path_cannot_be_empty(self):
        """Path cannot be empty"""
        with self.assertRaisesRegex(ConfigError, 'Path cannot be empty'):
            Directory('')

    def test_path_must_be_absolute(self):
        """Path must be absolute"""
        path = Path('foo', 'bar')
        with self.assertRaises(ConfigError) as cm:
            Directory(str(path))

        self.assertEqual(
            f'Path is not absolute: {path}',
            str(cm.exception).strip("'")
        )

    def test_path_must_be_directory(self):
        """Path must be a directory that exists"""
        path = Path(gettempdir(), 'foo').resolve()
        with self.assertRaises(ConfigError) as cm:
            Directory(path=str(path))

        self.assertEqual(
            f'Path is not a directory: {path}',
            str(cm.exception).strip("'")
        )

    def test_path_has_tilde(self):
        """Path can contain tilde (~)"""
        backup_path = Path.home() / 'backup'
        backup_path.mkdir(exist_ok=True)
        subject = Directory(str(backup_path))

        self.assertEqual(Path('~', 'backup').expanduser(), subject.path())

        backup_path.rmdir()

    def test_backup(self):
        """.backup()"""
        source_path, backup_path = self.create_file_fixtures(location='source')
        subject = Directory(path=str(backup_path))

        with self.assertLogs('clibato', None) as cm:
            subject.backup([
                Content(self.BUNNY_PATH, source_path / self.BUNNY_PATH),
                Content(self.WABBIT_PATH, source_path / self.WABBIT_PATH)
            ])

        self.assert_length(cm.records, 2)
        self.assert_log_record(
            cm.records[0],
            level='INFO',
            message=f'Backed up: {source_path / self.BUNNY_PATH}'
        )
        self.assert_log_record(
            cm.records[1],
            level='INFO',
            message=f'Backed up: {source_path / self.WABBIT_PATH}'
        )

        self.assert_file_contents(backup_path / self.BUNNY_PATH, 'I am a bunny')
        self.assert_file_contents(backup_path / self.WABBIT_PATH, 'I am a wabbit')

    def test_backup_file_not_found(self):
        """.backup() logs and continues if a file is not found"""
        source_path, backup_path = self.create_file_fixtures(location='source')
        skunk_path = '.skunk'
        subject = Directory(path=str(backup_path))

        with self.assertLogs('clibato', None) as cm:
            subject.backup([
                Content(skunk_path, source_path / skunk_path),
                Content(self.WABBIT_PATH, source_path / self.WABBIT_PATH)
            ])

        self.assert_length(cm.records, 2)
        self.assert_log_record(
            cm.records[0],
            level='ERROR',
            message=f"[Errno 2] No such file or directory: '{source_path / skunk_path}'"
        )
        self.assert_log_record(
            cm.records[1],
            level='INFO',
            message=f'Backed up: {source_path / self.WABBIT_PATH}'
        )

        self.assert_file_not_exists(backup_path / skunk_path)
        self.assert_file_contents(backup_path / self.WABBIT_PATH, 'I am a wabbit')

    def test_restore(self):
        """.restore()"""
        source_path, backup_path = self.create_file_fixtures(location='backup')
        subject = Directory(path=str(backup_path))

        with self.assertLogs('clibato', None) as cm:
            subject.restore([
                Content(self.BUNNY_PATH, source_path / self.BUNNY_PATH),
                Content(self.WABBIT_PATH, source_path / self.WABBIT_PATH)
            ])

        self.assert_length(cm.records, 2)
        self.assert_log_record(
            cm.records[0],
            level='INFO',
            message=f'Restored: {source_path / self.BUNNY_PATH}'
        )
        self.assert_log_record(
            cm.records[1],
            level='INFO',
            message=f'Restored: {source_path / self.WABBIT_PATH}'
        )

        self.assert_file_contents(source_path / self.BUNNY_PATH, 'I am a bunny')
        self.assert_file_contents(source_path / self.WABBIT_PATH, 'I am a wabbit')

    def test_restore_file_not_found(self):
        """.restore() logs and continues if a file is not found"""
        source_path, backup_path = self.create_file_fixtures(location='backup')
        skunk_path = '.skunk'
        subject = Directory(path=str(backup_path))

        with self.assertLogs('clibato', None) as cm:
            subject.restore([
                Content(skunk_path, source_path / skunk_path),
                Content(self.WABBIT_PATH, source_path / self.WABBIT_PATH)
            ])

        self.assert_length(cm.records, 2)
        self.assert_log_record(
            cm.records[0],
            level='ERROR',
            message=f"[Errno 2] No such file or directory: '{backup_path / skunk_path}'"
        )
        self.assert_log_record(
            cm.records[1],
            level='INFO',
            message=f'Restored: {source_path / self.WABBIT_PATH}'
        )

        self.assert_file_not_exists(source_path / skunk_path)
        self.assert_file_contents(source_path / self.WABBIT_PATH, 'I am a wabbit')


class TestRepository(TestCase):
    """Test destination.Repository"""

    def test_new(self):
        """Instance creation."""
        subject = Repository(
            gettempdir(),
            'git@github.com:jigarius/clibato.git',
            'backup',
            'Jigarius',
            'jigarius@example.com',
        )

        self.assertIsInstance(subject, Repository)

    def test__eq__(self):
        """__eq__()"""
        subject = Repository(gettempdir(), 'git@github.com:bunny/wabbit.git')

        self.assertEqual(
            subject,
            Repository(gettempdir(), 'git@github.com:bunny/wabbit.git')
        )

        self.assertNotEqual(
            subject,
            Repository(gettempdir(), 'git@github.com:bucky/wabbit.git')
        )

        self.assertNotEqual(
            subject,
            Directory(gettempdir())
        )

    def test_inheritance(self):
        """Repository must extend Directory"""
        self.assert_is_subclass(Repository, Destination)

    def test_path_cannot_be_empty(self):
        """Path cannot be empty"""
        message = 'Path cannot be empty'
        with self.assertRaisesRegex(ConfigError, message):
            Repository('', 'git@github.com:jigarius/clibato.git')

    def test_remote_cannot_be_empty(self):
        """Remote cannot be empty"""
        message = 'Remote cannot be empty'
        with self.assertRaisesRegex(ConfigError, message):
            Repository(gettempdir(), '')

    @unittest.skip('TODO')
    def test_backup(self):
        """.backup()"""

    @unittest.skip('TODO')
    def test_restore(self):
        """.backup()"""



import setuptools
from clibato import Clibato

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

GITHUB_URL = "https://github.com/jigarius/clibato"
AUTHOR_URL = "https://jigarius.com/"

setuptools.setup(
    name="clibato",
    version=Clibato.VERSION,
    author="Jigar Mehta",
    author_email="hello@jigarius.com",
    description="A tool to backup/restore simple files, e.g. dot files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='GNU LGPL v2.1',
    url=GITHUB_URL,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
    ],
    keywords='cli backup restore dotfile utility',
    project_urls={
        "Documentation": GITHUB_URL + '#readme',
        "Author": AUTHOR_URL,
        "Tracker": "https://github.com/jigarius/clibato/issues"
    },
    packages=setuptools.find_packages('clibato', exclude=['test']),
    python_requires='>=3.6',
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['clibato=clibato.__main__:main'],
    }
)



import unittest
from pathlib import Path
from clibato import Content, ConfigError


class TestContent(unittest.TestCase):
    """Test clibato.Content"""

    def test__eq__(self):
        """.__eq__()"""
        source_path = str(Path('~', 'todo.txt'))
        subject = Content('todo.txt', source_path)

        self.assertEqual(
            subject,
            Content('todo.txt', source_path)
        )

        self.assertNotEqual(
            subject,
            Content('done.txt', str(Path('~', 'done.txt')))
        )

    def test_from_dict(self):
        """.from_dict()"""
        self.assertEqual(
            [Content('.bashrc', None)],
            Content.from_dict({'.bashrc': None})
        )

    def test_from_dict_source_path_cannot_be_dictionary(self):
        """.from_dict() fails if content entry is a dictionary"""
        with self.assertRaisesRegex(ConfigError, 'Illegal source path for .bashrc: {}'):
            Content.from_dict({'.bashrc': {}})

    def test_backup_path(self):
        """.backup_path() works"""
        subject = Content('.bashrc')

        self.assertEqual(Path('.bashrc'), subject.backup_path())

    def test_source_path_with_absolute_path(self):
        """.new() works with absolute source paths"""
        source_path = Path('/', 'todo.txt').absolute()
        subject = Content('todo.txt', str(source_path))

        self.assertEqual(source_path, subject.source_path())

    def test_source_path_with_tilde(self):
        """.new() works with source paths starting with tilde"""
        source_path = Path('~', 'Documents', 'todo.txt')
        subject = Content('todo.txt', str(source_path))

        self.assertEqual(
            source_path.expanduser().absolute(),
            subject.source_path()
        )

    def test_source_path_with_relative_path(self):
        """.new() raises if source_path is relative"""
        path = Path('Documents', 'todo.txt')
        with self.assertRaises(ConfigError) as cm:
            Content('todo.txt', path)

        self.assertEqual(
            f'Source path invalid: {path}',
            str(cm.exception).strip("'")
        )

    def test_source_path_when_empty(self):
        """.new() works when source_path is empty"""
        subject = Content('.bashrc', '')

        self.assertEqual(
            Path('~', '.bashrc').expanduser(),
            subject.source_path()
        )

    def test_source_path_when_undefined(self):
        """.new() works when source_path is None"""
        subject = Content('.bashrc')

        self.assertEqual(
            Path('~', '.bashrc').expanduser(),
            subject.source_path()
        )

    def test_backup_path_cannot_be_absolute(self):
        """.new() raises if backup path is not absolute"""
        path = str(Path('/', 'Documents', '.bashrc').absolute())
        with self.assertRaises(ConfigError) as cm:
            Content(path)

        self.assertEqual(
            f'Backup path cannot be absolute: {path}',
            str(cm.exception).strip("'")
        )

    def test_backup_path_cannot_contain_illegal_elements(self):
        """.new() raises if backup path contains illegal elements"""
        illegal_parts = ['~', '..']

        for part in illegal_parts:
            message = f'Backup path cannot contain: {part}'
            with self.assertRaisesRegex(ConfigError, message):
                Content(str(Path(part, '.bashrc')))
